from datetime import timedelta

from data import dataset
from invoicing.duration import sum_dataset_durations


def calculate_acquisition_time(
    token: str, investigation_id: float, gap=20, dataset_type="acquisition"
) -> timedelta:
    datasets = dataset.get_datasets(token, investigation_ids=str(investigation_id), dataset_type=dataset_type)
    return sum_dataset_durations(datasets, gap)



import csv
import random
import statistics
import sys
import time
from collections import defaultdict
from functools import wraps

import typer
from icat_plus_client.models.session import Session

from data import dataset, investigation, sample, session

performance_app = typer.Typer(help="Performance workbench commands")


def get_role(user: Session):
    if user.is_administrator:
        return "administrator"
    if user.is_instrument_scientist:
        return "instrumentScientist"
    return "user"


# Decorator to measure execution time
def timed(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        duration = (end_time - start_time) * 1000
        return result, round(duration)

    return wrapper


@timed
def get_timed_datasets(**kwargs):
    return dataset.get_datasets(**kwargs)


@timed
def get_timed_samples(**kwargs):
    return sample.get_samples_by(**kwargs)


def get_timed_dataset_by_samples(**kwargs):
    return sample.get_samples_by(**kwargs)


def workbench(
    fetch_fn,
    result_label: str,
    *,
    random_seed: int = 42,
    cooldown_seconds: float = 0,
    skip_range=(0, 100),
    limit_range=(100, 200),
    x: int = 1,
    y: int = 1,
):
    """
    Generic benchmark decorator.

    fetch_fn: timed callable (get_timed_datasets / get_timed_samples)
    result_label: CSV column name ('datasets' / 'samples')
    """

    def decorator(func):
        @wraps(func)
        def wrapper(
            tokens: str = typer.Option(..., "-t", "--token"),
            investigation_ids: str | None = typer.Option(None, "-i", "--investigation-ids"),
            start_date: str = typer.Option(
                None,
                "-s",
                "--start-date",
                help="Start date range. Format is YYYY-MM-DD. Example: 2019-01-01",
            ),
            end_date: str = typer.Option(
                None,
                "-e",
                "--end-date",
                help="End date range. Format is YYYY-MM-DD. Example: 2019-01-01",
            ),
            skip_count: int | None = typer.Option(
                1,
                "-k",
                "--skip-count",
                help="Number of skip variations",
            ),
            limit_count: int | None = typer.Option(
                1,
                "-l",
                "--limit-count",
                help="Number of limit variations",
            ),
            output: str | None = typer.Option(None, "-o", "--output"),
        ):
            user_tokens = tokens.split(",")

            investigation_id_list = []
            if investigation_ids is not None:
                investigation_id_list = investigation_ids.split(",")
            else:
                investigations = investigation.get_investigation_by_id(
                    user_tokens[0], start_date=start_date, end_date=end_date
                )
                investigation_id_list = [i.id for i in investigations]

            fields = [
                "role",
                "user",
                "investigation_id",
                "skip",
                "limit",
                result_label,
                "time(ms)",
            ]

            stdout_writer = csv.DictWriter(sys.stdout, fieldnames=fields, delimiter="\t")
            stdout_writer.writeheader()

            file_writer = None
            f = None
            if output:
                f = open(output, "w", newline="")
                file_writer = csv.DictWriter(f, fieldnames=fields, delimiter="\t")
                file_writer.writeheader()

            role_timings = defaultdict(list)

            try:
                random.seed(random_seed)

                # --------------------------------------------------------------
                # Build randomized calls
                # --------------------------------------------------------------
                calls = []
                for investigation_id in investigation_id_list:
                    for _ in range(skip_count):
                        skip = str(random.randint(*skip_range))
                        for _ in range(limit_count):
                            limit = str(random.randint(*limit_range))
                            for token in user_tokens:
                                calls.append(
                                    {
                                        "token": token,
                                        "investigation_id": investigation_id,
                                        "skip": skip,
                                        "limit": limit,
                                    }
                                )

                random.shuffle(calls)

                # --------------------------------------------------------------
                # Execute benchmark
                # --------------------------------------------------------------
                for call in calls:
                    token = call["token"]
                    investigation_id = call["investigation_id"]
                    skip = call["skip"]
                    limit = call["limit"]

                    user_session: Session = session.get_info(token)
                    role = get_role(user_session.user)

                    result, duration = fetch_fn(
                        token=token,
                        investigation_ids=investigation_id,
                        limit=limit,
                        skip=skip,
                    )

                    role_timings[role].append(duration)

                    row = {
                        "role": role,
                        "user": user_session.user.username,
                        "investigation_id": investigation_id,
                        "skip": skip,
                        "limit": limit,
                        result_label: len(result),
                        "time(ms)": duration,
                    }

                    stdout_writer.writerow(row)
                    if file_writer:
                        file_writer.writerow(row)

                    time.sleep(cooldown_seconds)

            finally:
                if f:
                    f.close()

            # --------------------------------------------------------------
            # Statistics
            # --------------------------------------------------------------
            print("\n=== Benchmark statistics per role ===\n")
            for role, times in role_timings.items():
                if len(times) < 2:
                    continue
                print(
                    f"Role: {role} | "
                    f"count={len(times)} | "
                    f"avg={statistics.mean(times):.2f}ms | "
                    f"min={min(times):.2f}ms | "
                    f"max={max(times):.2f}ms | "
                    f"std={statistics.stdev(times):.2f}ms"
                )

        return wrapper

    return decorator


@performance_app.command("dataset")
@workbench(
    fetch_fn=get_timed_datasets,
    result_label="datasets",
)
def dataset_workbench(
    tokens: str = typer.Option(..., "-t", "--token", help="Comma separated list of ICAT tokens"),
    investigation_ids: str | None = typer.Option(
        None,
        "--investigation-ids",
        "-i",
        help="Comma separated list of investigation IDs (optional)",
    ),
    start_date: str = typer.Option(
        None,
        "-s",
        "--start-date",
        help="Start date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    end_date: str = typer.Option(
        None,
        "-e",
        "--end-date",
        help="End date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    skip_count: int | None = typer.Option(
        1,
        "-k",
        "--skip-count",
        help="Number of skip variations",
    ),
    limit_count: int | None = typer.Option(
        1,
        "-l",
        "--limit-count",
        help="Number of limit variations",
    ),
    output: str | None = typer.Option(
        None,
        "-o",
        "--output",
        help="CSV output file path. Example: myfile.csv",
    ),
):
    """Benchmark dataset retrieval."""


@performance_app.command(
    "sample",
    help="""
    Generates volume statistics for all instruments over a date range.

    Example usage:

    .. code-block:: bash

      minedrac performance sample --token <TOKEN>  --start-date 2019-02-01 --end-date 2019-02-02

    .. code-block:: bash

      === Benchmark statistics per role ===

        Role: instrumentScientist | count=28 | avg=155.68ms | min=62.00ms | max=856.00ms | std=217.77ms
        Role: administrator | count=28 | avg=169.14ms | min=64.00ms | max=583.00ms | std=180.10ms
        Role: user | count=28 | avg=117.39ms | min=66.00ms | max=560.00ms | std=125.73ms
                               """,
)
@workbench(
    fetch_fn=get_timed_samples,
    result_label="samples",
)
def samples_workbench(
    tokens: str = typer.Option(..., "-t", "--token", help="Comma separated list of ICAT tokens"),
    investigation_ids: str | None = typer.Option(
        None,
        "--investigation-ids",
        "-i",
        help="Comma separated list of investigation IDs (optional)",
    ),
    start_date: str = typer.Option(
        None,
        "-s",
        "--start-date",
        help="Start date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    end_date: str = typer.Option(
        None,
        "-e",
        "--end-date",
        help="End date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    skip_count: int | None = typer.Option(
        1,
        "-k",
        "--skip-count",
        help="Number of skip variations",
    ),
    limit_count: int | None = typer.Option(
        1,
        "-l",
        "--limit-count",
        help="Number of limit variations",
    ),
    output: str | None = typer.Option(
        None,
        "-o",
        "--output",
        help="CSV output file path. Example: myfile.csv",
    ),
):
    """Benchmark dataset retrieval."""


# For click documentation
click_performance_app = typer.main.get_command(performance_app)



import functools
import time
from collections.abc import Callable
from typing import Any


def timed(func: Callable[..., Any]):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        result = func(*args, **kwargs)
        duration = time.perf_counter() - start
        return result, duration

    return wrapper



import json

import pandas as pd
import typer

from data import session
from stats.technique import technique

technique_statistics_app = typer.Typer(help="Technique related commands for statistics")


def validate_token(token: str):
    session_response = session.get_info(token)
    if session_response is None:
        typer.echo(" Invalid or expired token. Exiting.")
        raise typer.Exit(code=1)


@technique_statistics_app.command("investigation", help="Gets technique statistics per investigation")
def statistics_technique_investigation(
    token: str = typer.Option(..., "-t", "--token", help="ICAT token"),
    start_date: str = typer.Option(
        ...,
        "-s",
        "--start_date",
        help="Start date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    end_date: str = typer.Option(
        ...,
        "-e",
        "--end_date",
        help="End date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    output_format: str = typer.Option(
        "csv",
        "-f",
        "--format",
        help="Output format: csv | json | stdout",
    ),
    output_file: str = typer.Option(
        None,
        "-o",
        "--output",
        help="Output filepath",
    ),
):
    validate_token(token)
    data = technique.get_techniques_by_investigation_cli(token, start_date, end_date)
    if output_format == "csv":
        if not output_file:
            raise typer.BadParameter("You must provide --output when format=csv")
        technique.write_csv(data, output_file)
    elif output_format == "json":
        if output_file:
            with open(output_file, "w") as f:
                json.dump(data, f, indent=2)
        else:
            print(json.dumps(data, indent=2))

    elif output_format == "stdout":
        df = pd.DataFrame(data)
        print(df)

    else:
        raise typer.BadParameter("Invalid format. csv | json | stdout")


@technique_statistics_app.command("beamline", help="Gets technique statistics per beamline")
def statistics_technique_beamline(
    token: str = typer.Option(..., "-t", "--token", help="ICAT token"),
    start_date: str = typer.Option(
        ...,
        "-s",
        "--start_date",
        help="Start date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    end_date: str = typer.Option(
        ...,
        "-e",
        "--end_date",
        help="End date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    output_format: str = typer.Option(
        "csv",
        "-f",
        "--format",
        help="Output format: csv | json | stdout",
    ),
    output_file: str = typer.Option(
        None,
        "-o",
        "--output",
        help="Output filepath",
    ),
):
    validate_token(token)
    data = technique.get_techniques_by_beamlines(token, start_date, end_date)
    if output_format == "csv":
        if not output_file:
            raise typer.BadParameter("You must provide --output when format=csv")
        technique.write_csv(data, output_file)
    elif output_format == "json":
        if output_file:
            with open(output_file, "w") as f:
                json.dump(data, f, indent=2)
        else:
            print(json.dumps(data, indent=2))

    elif output_format == "stdout":
        df = pd.DataFrame(data)
        print(df)

    else:
        raise typer.BadParameter("Invalid format. csv | json | stdout")


click_technique_statistics_app = typer.main.get_command(technique_statistics_app)



from datetime import timedelta

import pytest
from icat_plus_client.models.dataset import Dataset
from icat_plus_client.models.dataset_document_investigation import (
    DatasetDocumentInvestigation,
)

from invoicing.duration import sum_dataset_durations


# Fixture for investigation
@pytest.fixture
def investigation():
    return DatasetDocumentInvestigation(
        id=1,
        name="INV",
        startDate="2024-01-30T10:00:00.000Z",
        endDate="2024-01-30T12:00:00.000Z",
        location="/tmp",
        sampleName="",
    )


@pytest.fixture
def make_datasets(investigation):
    def _make(*dataset_times):
        return [
            Dataset(
                id=i,
                name=f"dataset-{i}",
                start_date=start,
                end_date=end,
                location="/tmp",
                sample_name="",
                parameters=[],
                investigation=investigation,  # Pydantic requirement
            )
            for i, (start, end) in enumerate(dataset_times, start=1)
        ]

    return _make


@pytest.mark.parametrize(
    "dataset_times, merge_gap_minutes, expected, ignore_longer_than",
    [
        pytest.param(
            [("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z")],
            20,
            timedelta(hours=1),
            5,
            id="Single dataset duration with 1h dataset",
        ),
        pytest.param(
            [("2024-01-30T10:00:00.000Z", "2024-01-30T10:00:01.000Z")],
            20,
            timedelta(seconds=1),
            5,
            id="Single dataset duration with 1s dataset",
        ),
        pytest.param(
            [("2024-01-30T10:00:00.000Z", "2024-01-30T10:00:00.000Z")],
            20,
            timedelta(seconds=0),
            5,
            id="Single dataset duration with 0s dataset",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T11:00:00.000Z", "2024-01-30T12:00:00.000Z"),
            ],
            20,
            timedelta(hours=2, minutes=0),
            5,
            id="Two datasets no gap in between",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T11:30:00.000Z", "2024-01-30T12:00:00.000Z"),
            ],
            20,
            timedelta(hours=1, minutes=30),
            5,
            id="Two datasets with gap in between",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T11:30:00.000Z", "2024-01-30T12:00:00.000Z"),
            ],
            30,
            timedelta(hours=2, minutes=0),
            5,
            id="Two datasets with gap short enough in between",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
            ],
            30,
            timedelta(hours=1, minutes=0),
            5,
            id="Two fully overlaping datasets with gap short enough in between",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T10:30:00.000Z", "2024-01-30T11:30:00.000Z"),
            ],
            30,
            timedelta(hours=1, minutes=30),
            5,
            id="Two partial overlaping datasets with gap short enough in between",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-29T10:30:00.000Z", "2024-01-29T11:30:00.000Z"),
            ],
            30,
            timedelta(hours=2, minutes=0),
            5,
            id="Two partial overlaping datasets in different days",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T11:30:00.000Z", "2024-01-30T12:00:00.000Z"),
                ("2024-01-30T12:30:00.000Z", "2024-01-30T13:00:00.000Z"),
            ],
            30,
            timedelta(hours=3, minutes=0),
            5,
            id="Three datasets with no gaps",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T11:00:00.000Z"),
                ("2024-01-30T11:30:00.000Z", "2024-01-30T12:00:00.000Z"),
                ("2024-01-30T12:30:00.000Z", "2024-01-30T13:00:00.000Z"),
                ("2024-01-30T12:30:00.000Z", "2024-01-30T22:00:00.000Z"),
            ],
            30,
            timedelta(hours=3, minutes=0),
            5,
            id="Three datasets with no gaps and one to be ignored",
        ),
        pytest.param(
            [
                ("2024-01-30T10:00:00.000Z", "2024-01-30T12:00:00.000Z"),
                ("2024-01-30T12:30:00.000Z", "2024-01-30T22:00:00.000Z"),
            ],
            30,
            timedelta(hours=0, minutes=0),
            1,
            id="All ignored",
        ),
    ],
)
def test_calculate_dataset_duration(
    dataset_times, merge_gap_minutes, expected, ignore_longer_than, make_datasets
):
    """
    Runs the merge-gap test with multiple dataset inputs.
    """
    datasets = make_datasets(*dataset_times)

    result = sum_dataset_durations(
        datasets, merge_gap_minutes=merge_gap_minutes, outfile=None, ignore_longer_than=ignore_longer_than
    )
    assert result == expected



import typer

from authentication.cli import authentication_app
from invoicing.cli import invoicing_app
from performance.cli import performance_app
from stats.cli import statistics_app

app = typer.Typer(help="ICAT+ CLI")

# Sub-Typer for commands
login_app = typer.Typer(help="Log in command")


config_app = typer.Typer(help="Minedrac Configuration Settings")

app.add_typer(login_app, name="login")
app.add_typer(config_app, name="config")
app.add_typer(invoicing_app, name="invoicing")
app.add_typer(authentication_app, name="login")
app.add_typer(performance_app, name="performance")
app.add_typer(statistics_app, name="statistics")


def main():
    app()


if __name__ == "__main__":
    main()

click_app = typer.main.get_command(app)



import datetime


def format_date(date_str: str) -> str:
    """Convert ISO date string to YYYY-MM-dd-hh:mm."""
    try:
        dt = datetime.fromisoformat(date_str.replace("Z", "+00:00"))  # handle UTC 'Z' if present
        return dt.strftime("%Y-%m-%d-%H:%M")
    except ValueError:
        return ""



import csv
import logging
from io import StringIO


class CSVFormatter(logging.Formatter):
    def __init__(self, fieldnames, fmt="%(asctime)s [%(levelname)s] %(name)s: %(message)s"):
        super().__init__(fmt)
        self.fieldnames = fieldnames

    def format(self, record):
        # Convert the extra fields into a CSV line
        row = {field: getattr(record, field, "") for field in self.fieldnames}
        output = StringIO()
        writer = csv.DictWriter(output, fieldnames=self.fieldnames, extrasaction="ignore", lineterminator="")
        writer.writerow(row)
        # Use CSV line as the actual message
        record.msg = output.getvalue()
        return super().format(record)



import csv
import sys
from datetime import datetime

import typer

from data import dataset, investigation
from data.parcel import get_declared_samples_by, get_parcels_by
from invoicing import duration
from stats.volume.volume import get_volumes_by_investigation

invoicing_app = typer.Typer(help="Invoicing commands")


@invoicing_app.command("report", help="Generates a report for invoicong for each investigation")
def invoicing_report(
    token: str = typer.Option(..., "-t", "--token", help="ICAT session ID"),
    investigation_ids: str | None = typer.Option(
        None,
        "--investigation-ids",
        "-i",
        help="Comma separated list of investigation IDs (optional)",
    ),
    instrument_name: str | None = typer.Option(None, help="Filter by instrument name"),
    investigation_name: str | None = typer.Option(None, help="Filter by instrument name"),
    start_date: str = typer.Option(
        None,
        "-s",
        "--start-date",
        help="Start date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    end_date: str = typer.Option(
        None,
        "-e",
        "--end-date",
        help="End date range. Format is YYYY-MM-DD. Example: 2019-01-01",
    ),
    unit: str = typer.Option(
        "GB",
        "-u",
        "--unit",
        help="It can be MB, GB or TB (otherwise will return in GB)",
    ),
    merge_gap_minutes: int = typer.Option(
        20,
        "-g",
        "--gap",
        help="Gap in minutes to merge datasets. Less than 'gap' minutes the datasets time will be merged",
    ),
    output: str = typer.Option(
        None,
        "-o",
        "--output",
        help="CSV output file path. Example: myfile.csv",
    ),
    log: str = typer.Option(
        None,
        "-l",
        "--log",
        help="CSV log output file path. Example: myfile_log.csv",
    ),
    ignore_longer_than: int = typer.Option(
        None,
        "-ig",
        "--ignore",
        help="The time in hours that will be taken into account to omit a dataset as considered an error.",
    ),
):
    investigations = investigation.get_investigation_by_id(
        session_id=token,
        investigation_id=investigation_ids,
        start_date=start_date,
        end_date=end_date,
        instrument_name=instrument_name,
        investigation_name=investigation_name,
    )

    fields = [
        "proposal",
        "investigation_id",
        "beamline",
        "start_date",
        "end_date",
        "booked_time",
        "samples_collected",
        "samples_declared",
        "parcels",
        "volume",
        "raw_volume",
        "processed_volume",
        "effective_time",
        "total_time",
    ]

    writers = [csv.DictWriter(sys.stdout, fieldnames=fields, delimiter="\t")]
    # If file provided, also write to the file
    if output:
        f = open(output, "w", newline="")
        writers.append(csv.DictWriter(f, fieldnames=fields, delimiter="\t"))
    else:
        f = None

    # Write headers on all writers
    for w in writers:
        w.writeheader()

    for inv in investigations:
        (total, raw, processed) = get_volumes_by_investigation([inv], unit)
        samples_declared = get_declared_samples_by(token, str(inv.id))
        parcels = get_parcels_by(token, str(inv.id))
        start = datetime.fromisoformat(inv.start_date)
        end = datetime.fromisoformat(inv.end_date) if inv.end_date is not None else start
        row = {
            "proposal": inv.name,
            "investigation_id": inv.id,
            "beamline": inv.instrument.name,
            "start_date": start.strftime("%Y-%m-%d %H:%M:%S"),
            "end_date": end.strftime("%Y-%m-%d %H:%M:%S"),
            "booked_time": end - start,
            "volume": f"{total} {unit}",
            "raw_volume": f"{raw} {unit}",
            "processed_volume": f"{processed} {unit}",
        }

        """Calculate total dataset duration for a session."""
        datasets = dataset.get_datasets(
            session_id=token,
            investigation_ids=str(inv.id),
            dataset_type="acquisition",
        )

        samples_collected = sorted({d.sample_name for d in datasets if getattr(d, "sample_name", None)})
        row["samples_collected"] = len(samples_collected)
        row["samples_declared"] = len(samples_declared)
        row["parcels"] = len(parcels)
        row["total_time"] = duration.sum_dataset_durations(
            datasets, merge_gap_minutes, log, ignore_longer_than
        )
        row["effective_time"] = duration.sum_dataset_durations(datasets, 0, None, ignore_longer_than)

        for w in writers:
            w.writerow(row)


click_invoicing_app = typer.main.get_command(invoicing_app)



import logging

import icat_plus_client
from dotenv import load_dotenv
from icat_plus_client.models.instrument import Instrument

from config import configuration

# Load .env if needed
load_dotenv()

# Setup logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_instruments() -> list[Instrument]:
    """Fetch all instruments from ICAT Catalogue."""
    with icat_plus_client.ApiClient(configuration) as api_client:
        api_instance = icat_plus_client.CatalogueApi(api_client)
        try:
            instruments: list[Instrument] = api_instance.catalogue_instruments_get()

            if instruments:
                logger.debug("Fetched %d instruments successfully.", len(instruments))
                return instruments
            else:
                logger.warning("No instruments found in the catalogue.")
                return []

        except icat_plus_client.rest.ApiException as api_exc:
            logger.error(
                "API Exception when calling CatalogueApi->catalogue_instruments_get: %s",
                api_exc,
            )
        except Exception as e:
            logger.exception("Unexpected error while fetching instruments: %s", e)

        return []



from data import sample


def count_samples_with_datasets(token: str, investigation_id: float) -> int:
    samples = sample.get_samples_by(token, investigation_id)
    count = 0
    for s in samples:
        for param in s.parameters:
            if param.name == "__datasetCount" and int(param.value) > 0:
                count += 1
                break  # no need to check other params for this sample
    return count



import csv
import sys
from collections.abc import Callable
from functools import wraps

from icat_plus_client.models.investigation import Investigation

from data import instrument, investigation
from stats import investigation_parameters


def get_volumes_by_investigation(
    investigations: list[Investigation], unit: str = "bytes"
) -> tuple[float, float, float]:
    """
    Returns the total volumes (__volume, __acquisitionVolume, __processedVolume)
    converted into the desired unit, truncated to 2 decimal places.

    Args:
        investigations (list): List of investigations.
        unit (str): "bytes", "MB", "GB", or "TB". Default is "bytes".

    Returns:
        tuple(float, float, float): Volumes in the requested unit.
    """
    conversion = {
        "bytes": 1,
        "MB": 1024**2,
        "GB": 1024**3,
        "TB": 1024**4,
    }

    if unit not in conversion:
        raise ValueError(f"Invalid unit '{unit}'. Use one of {list(conversion.keys())}.")

    def truncate(number: float, digits: int = 2) -> float:
        stepper = 10.0**digits
        return int(number * stepper) / stepper

    if investigations is not None:
        __volume = investigation_parameters.sum_investigation_parameter(investigations, "__volume")
        __acquisitionVolume = investigation_parameters.sum_investigation_parameter(
            investigations, "__acquisitionVolume"
        )
        __processedVolume = investigation_parameters.sum_investigation_parameter(
            investigations, "__processedVolume"
        )

        factor = conversion[unit]
        return (
            int(truncate(__volume / factor, 2)),
            int(truncate(__acquisitionVolume / factor, 2)),
            int(truncate(__processedVolume / factor, 2)),
        )

    return 0.0, 0.0, 0.0


def yearly_csv(fields: list[str] | None = None, delimiter: str = "\t"):
    """
    Decorator to generate CSV per year.
    The decorated function should have signature:
        func(token: str, year: int, unit: str) -> Dict[str, float]
    """

    def decorator(func: Callable[[str, int, str], dict[str, float]]):
        @wraps(func)
        def wrapper(token: str, start_year: int, end_year: int, unit: str = ""):
            writer = csv.DictWriter(sys.stdout, fieldnames=fields, delimiter=delimiter)
            writer.writeheader()

            for year in range(start_year, end_year + 1):
                row = func(token, year, unit)
                writer.writerow(row)

        return wrapper

    return decorator


def get_datasets_count_by(investigations: list[Investigation]) -> float:
    return investigation_parameters.sum_investigation_parameter(investigations, "__datasetCount")


@yearly_csv(fields=["year", "dataset_count"])
def get_datasets_by_year(token: str, year: int, unit: str) -> dict[str, float]:
    investigations: list[Investigation] = investigation.get_investigation_by_id(
        session_id=token,
        start_date=f"{year}-01-01",
        end_date=f"{year}-12-31",
    )
    dataset_count = get_datasets_count_by(investigations=investigations)
    return {
        "year": year,
        "dataset_count": dataset_count,
    }


@yearly_csv(fields=["year", "total_volume", "raw_volume", "processed_volume"])
def get_volume_by_year(token: str, year: int, unit: str) -> dict[str, float]:
    investigations = investigation.get_investigation_by_id(
        session_id=token,
        start_date=f"{year}-01-01",
        end_date=f"{year}-12-31",
    )
    total_volume, raw_volume, processed_volume = get_volumes_by_investigation(
        investigations=investigations, unit=unit
    )
    return {
        "year": year,
        "total_volume": total_volume,
        "raw_volume": raw_volume,
        "processed_volume": processed_volume,
    }


def get_volume_by_beamlines(token: str, start_date: str, end_date: str, unit: str, output: str):
    instruments = instrument.get_instruments()
    fields = [
        "beamline",
        "total_volume",
        "raw_volume",
        "processed_volume",
    ]
    writer = csv.DictWriter(sys.stdout, fieldnames=fields, delimiter="\t")
    writer.writeheader()

    file_writer = None
    f = None
    if output:
        f = open(output, "w", newline="")
        file_writer = csv.DictWriter(f, fieldnames=fields, delimiter="\t")
        file_writer.writeheader()

    for ins in instruments:
        investigations = investigation.get_investigation_by_id(
            session_id=token,
            instrument_name=ins.name,
            start_date=start_date,
            end_date=end_date,
        )
        [total_volume, acquisition_volume, processed_volume] = get_volumes_by_investigation(
            investigations=investigations, unit=unit
        )

        row = {
            "beamline": ins.name,
            "total_volume": total_volume,
            "raw_volume": acquisition_volume,
            "processed_volume": processed_volume,
        }

        writer.writerow(
            {
                "beamline": ins.name,
                "total_volume": total_volume,
                "raw_volume": acquisition_volume,
                "processed_volume": processed_volume,
            }
        )
        if file_writer:
            file_writer.writerow(row)


def get_volume_by_instrument_year(token: str, start_year: int, end_year: int, unit: str, output: str):
    instruments = instrument.get_instruments()
    # Build CSV fields: first "year", then each instrument's name
    fields = ["year"] + [ins.name for ins in instruments]

    writer = csv.DictWriter(sys.stdout, fieldnames=fields, delimiter="\t")
    writer.writeheader()

    for year in range(start_year, end_year + 1):
        row = {"year": year}  # Start with the year
        for ins in instruments:
            # Get all investigations in the date range
            investigations = investigation.get_investigation_by_id(
                session_id=token,
                start_date=f"{year}-01-01",
                end_date=f"{year}-12-31",
                instrument_name=ins.name,
            )
            [total_volume, acquisition_volume, processed_volume] = get_volumes_by_investigation(
                investigations=investigations, unit=unit
            )
            # Add total_volume to the row under the instrument name
            row[ins.name] = total_volume
        writer.writerow(row)



import logging

import icat_plus_client
from dotenv import load_dotenv
from icat_plus_client.models.sample import Sample

from config import configuration

# Load .env if needed
load_dotenv()

# Setup logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_samples_by(
    token: str,
    investigation_ids: str | None = None,
    limit: int | None = None,
    skip: int | None = None,
) -> list[Sample]:
    with icat_plus_client.ApiClient(configuration) as api_client:
        # Create an instance of the API class
        api_instance = icat_plus_client.CatalogueApi(api_client)
        try:
            # Gets samples
            samples = api_instance.catalogue_session_id_samples_get(
                session_id=token,
                investigation_id=str(investigation_ids),
                limit=limit,
                skip=skip,
                sort_by="DATE",
            )
            if samples:
                logger.debug("Fetched %d samples successfully.", len(samples))
                return samples
            else:
                logger.debug("No samples found for the given parameters.")
                return []
        except Exception as e:
            print(f"Exception when CatalogueApi->catalogue_session_id_samples_get: {e}")



import logging

import icat_plus_client
from dotenv import load_dotenv
from icat_plus_client.models.dataset import Dataset

from config import configuration

# Load .env if needed
load_dotenv()

# Setup logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


def get_datasets(
    token: str,
    investigation_ids: str | None = None,
    instrument_name: str | None = None,
    dataset_ids: str | None = None,
    dataset_type: str | None = None,
    nested: str | None = None,
    sample_id: str | None = None,
    limit: int | None = None,
    skip: int | None = None,
    search: str | None = None,
    sort_by: str | None = None,
    sort_order: str | None = None,
) -> list[Dataset]:
    with icat_plus_client.ApiClient(configuration) as api_client:
        api_instance = icat_plus_client.CatalogueApi(api_client)
        try:
            datasets: list[Dataset] = api_instance.catalogue_session_id_dataset_get(
                token,
                investigation_ids=investigation_ids,
                dataset_ids=dataset_ids,
                sample_id=sample_id,
                instrument_name=instrument_name,
                limit=limit,
                skip=skip,
                search=search,
                dataset_type=dataset_type,
                nested=nested,
                sort_by=sort_by,
                sort_order=sort_order,
            )

            if datasets:
                logger.debug("Fetched %d datasets successfully.", len(datasets))
                return datasets
            else:
                logger.debug("No datasets found for the given parameters.")
                return []

        except icat_plus_client.rest.ApiException as api_exc:
            logger.error(
                "API Exception when CatalogueApi->catalogue_session_id_dataset_get: %s",
                api_exc,
            )
        except Exception as e:
            logger.exception("Unexpected error while fetching datasets: %s", e)

        return []



import typer

from data import session

authentication_app = typer.Typer(help="Authentication commands")


@authentication_app.command("authenticate", help="Gets the ICAT's token based on username and password")
def authenticate_cli(
    authenticator: str = typer.Option(
        "db",
        "-a",
        "--authenticator",
        help="Plugin authentication. it can be db or esrf. Default is db",
    ),
    username: str = typer.Option(..., "-u", "--username", help="ICAT username"),
    password: str = typer.Option(..., "-p", "--password", help="ICAT password"),
):
    """Retrieve investigations by session_id."""
    session_response = session.get_session(authenticator, username, password)
    typer.echo(session_response)


@authentication_app.command("info", help="Gets the ICAT's session information based on a valid ICAT's token")
def get_session_info(
    token: str = typer.Option(..., "-t", "--token", help="ICAT token"),
):
    """Retrieve investigations by session_id."""
    session_response = session.get_info(token)
    typer.echo(session_response)
